/*========================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
                                 last edited: Dec.  14, 2009
========================================================================*/

#include "gpoplibc.h"

/*========================================================================
     gpop library classes
========================================================================*/

/*------------------------------------------------------------------------
     Tabulated informations on elements and methods
------------------------------------------------------------------------*/

/*----- tableOfElements ------------------------------------------------*/

tableOfElements::tableOfElements() {
  elemInfo e;
  clear();
  /* --- atomic weights are from Gaussian 03 source file utilam.F --- */
  e.aNbr =  1; e.aSym =  "H"; e.aWgt = 1.007825037;  push_back(e);
  e.aNbr =  2; e.aSym = "He"; e.aWgt = 4.00260325;   push_back(e);
  e.aNbr =  3; e.aSym = "Li"; e.aWgt = 7.0160045;    push_back(e);
  e.aNbr =  4; e.aSym = "Be"; e.aWgt = 9.0121825;    push_back(e);
  e.aNbr =  5; e.aSym =  "B"; e.aWgt = 11.0093053;   push_back(e);
  e.aNbr =  6; e.aSym =  "C"; e.aWgt = 12.0;         push_back(e);
  e.aNbr =  7; e.aSym =  "N"; e.aWgt = 14.003074008; push_back(e);
  e.aNbr =  8; e.aSym =  "O"; e.aWgt = 15.99491464;  push_back(e);
  e.aNbr =  9; e.aSym =  "F"; e.aWgt = 18.99840325;  push_back(e);
  e.aNbr = 10; e.aSym = "Ne"; e.aWgt = 19.9924391;   push_back(e);
  e.aNbr = 11; e.aSym = "Na"; e.aWgt = 22.9897697;   push_back(e);
  e.aNbr = 12; e.aSym = "Mg"; e.aWgt = 23.985045;    push_back(e);
  e.aNbr = 13; e.aSym = "Al"; e.aWgt = 26.9815413;   push_back(e);
  e.aNbr = 14; e.aSym = "Si"; e.aWgt = 27.9769284;   push_back(e);
  e.aNbr = 15; e.aSym =  "P"; e.aWgt = 30.9737634;   push_back(e);
  e.aNbr = 16; e.aSym =  "S"; e.aWgt = 31.9720718;   push_back(e);
  e.aNbr = 17; e.aSym = "Cl"; e.aWgt = 34.968852729; push_back(e);
  e.aNbr = 18; e.aSym = "Ar"; e.aWgt = 39.9623831;   push_back(e);
  e.aNbr = 19; e.aSym =  "K"; e.aWgt = 38.9637079;   push_back(e);
  e.aNbr = 20; e.aSym = "Ca"; e.aWgt = 39.9625907;   push_back(e);
  e.aNbr = 21; e.aSym = "Sc"; e.aWgt = 44.9559136;   push_back(e);
  e.aNbr = 22; e.aSym = "Ti"; e.aWgt = 47.9479467;   push_back(e);
  e.aNbr = 23; e.aSym =  "V"; e.aWgt = 50.9439625;   push_back(e);
  e.aNbr = 24; e.aSym = "Cr"; e.aWgt = 51.9405097;   push_back(e);
  e.aNbr = 25; e.aSym = "Mn"; e.aWgt = 54.9380463;   push_back(e);
  e.aNbr = 26; e.aSym = "Fe"; e.aWgt = 55.9349393;   push_back(e);
  e.aNbr = 27; e.aSym = "Co"; e.aWgt = 58.9331978;   push_back(e);
  e.aNbr = 28; e.aSym = "Ni"; e.aWgt = 57.9353471;   push_back(e);
  e.aNbr = 29; e.aSym = "Cu"; e.aWgt = 62.9295992;   push_back(e);
  e.aNbr = 30; e.aSym = "Zn"; e.aWgt = 63.9291454;   push_back(e);
  e.aNbr = 31; e.aSym = "Ga"; e.aWgt = 68.9255809;   push_back(e);
  e.aNbr = 32; e.aSym = "Ge"; e.aWgt = 73.9211788;   push_back(e);
  e.aNbr = 33; e.aSym = "As"; e.aWgt = 74.9215955;   push_back(e);
  e.aNbr = 34; e.aSym = "Se"; e.aWgt = 79.9165205;   push_back(e);
  e.aNbr = 35; e.aSym = "Br"; e.aWgt = 78.9183361;   push_back(e);
  e.aNbr = 36; e.aSym = "Kr"; e.aWgt = 83.9115064;   push_back(e);
  e.aNbr = 37; e.aSym = "Rb"; e.aWgt = 84.9117;      push_back(e);
  e.aNbr = 38; e.aSym = "Sr"; e.aWgt = 87.9056;      push_back(e);
  e.aNbr = 39; e.aSym =  "Y"; e.aWgt = 88.9054;      push_back(e);
  e.aNbr = 40; e.aSym = "Zr"; e.aWgt = 89.9043;      push_back(e);
  e.aNbr = 41; e.aSym = "Nb"; e.aWgt = 92.906;       push_back(e);
  e.aNbr = 42; e.aSym = "Mo"; e.aWgt = 97.9055;      push_back(e);
  e.aNbr = 43; e.aSym = "Tc"; e.aWgt = 98.9063;      push_back(e);
  e.aNbr = 44; e.aSym = "Ru"; e.aWgt = 101.9037;     push_back(e);
  e.aNbr = 45; e.aSym = "Rh"; e.aWgt = 102.9048;     push_back(e);
  e.aNbr = 46; e.aSym = "Pd"; e.aWgt = 105.9032;     push_back(e);
  e.aNbr = 47; e.aSym = "Ag"; e.aWgt = 106.90509;    push_back(e);
  e.aNbr = 48; e.aSym = "Cd"; e.aWgt = 113.9036;     push_back(e);
  e.aNbr = 49; e.aSym = "In"; e.aWgt = 114.9041;     push_back(e);
  e.aNbr = 50; e.aSym = "Sn"; e.aWgt = 117.9018;     push_back(e);
  e.aNbr = 51; e.aSym = "Sb"; e.aWgt = 120.9038;     push_back(e);
  e.aNbr = 52; e.aSym = "Te"; e.aWgt = 129.9067;     push_back(e);
  e.aNbr = 53; e.aSym =  "I"; e.aWgt = 126.9004;     push_back(e);
  e.aNbr = 54; e.aSym = "Xe"; e.aWgt = 131.9042;     push_back(e);
  e.aNbr = 55; e.aSym = "Cs"; e.aWgt = 133.9051;     push_back(e);
  e.aNbr = 56; e.aSym = "Ba"; e.aWgt = 137.905;      push_back(e);
/*
  e.aNbr = 1; e.aSym = "H"; e.aWgt = 1.007825; push_back(e);
  e.aNbr = 2; e.aSym = "He"; e.aWgt = 4.00260; push_back(e);
  e.aNbr = 3; e.aSym = "Li"; e.aWgt = 7.0150; push_back(e);
  e.aNbr = 4; e.aSym = "Be"; e.aWgt = 9.012194; push_back(e);
  e.aNbr = 5; e.aSym = "B"; e.aWgt = 11.009307; push_back(e);
  e.aNbr = 6; e.aSym = "C"; e.aWgt = 12.000000; push_back(e);
  e.aNbr = 7; e.aSym = "N"; e.aWgt = 14.003073; push_back(e);
  e.aNbr = 8; e.aSym = "O"; e.aWgt = 15.994914; push_back(e);
  e.aNbr = 9; e.aSym = "F"; e.aWgt = 18.99839; push_back(e);
  e.aNbr = 10; e.aSym = "Ne"; e.aWgt = 19.992441; push_back(e);
  e.aNbr = 11; e.aSym = "Na"; e.aWgt = 22.989761; push_back(e);
  e.aNbr = 12; e.aSym = "Mg"; e.aWgt = 23.985043; push_back(e);
  e.aNbr = 13; e.aSym = "Al"; e.aWgt = 26.981532; push_back(e);
  e.aNbr = 14; e.aSym = "Si"; e.aWgt = 27.976930; push_back(e);
  e.aNbr = 15; e.aSym = "P"; e.aWgt = 30.973764; push_back(e);
  e.aNbr = 16; e.aSym = "S"; e.aWgt = 31.972072; push_back(e);
  e.aNbr = 17; e.aSym = "Cl"; e.aWgt = 34.968853; push_back(e);
  e.aNbr = 18; e.aSym = "Ar"; e.aWgt = 39.962386; push_back(e);
  e.aNbr = 19; e.aSym = "K"; e.aWgt = 38.963711; push_back(e);
  e.aNbr = 20; e.aSym = "Ca"; e.aWgt = 39.962586; push_back(e);
*/
}

int tableOfElements::getAtmcNbr(string &sm) {
  string smR = sm;
  reformAtomicSymbol(smR);
  for (iterator p = begin(); p != end(); p++) {
    if (smR == p->aSym) { return p->aNbr; }
  }
  return 0;
}

double tableOfElements::getAtmcWeight(int anbr) {
  for (iterator p = begin(); p != end(); p++) {
    if (anbr == p->aNbr) { return p->aWgt; }
  }
  return 0.;
}

void tableOfElements::setAtmcWeight(int anbr, double wgt) {
  for (iterator p = begin(); p != end(); p++) {
    if (anbr == p->aNbr) { p->aWgt = wgt; }
  }
}

string tableOfElements::getAtmcSym(int anbr) {
  for (iterator p = begin(); p != end(); p++) {
    if (anbr == p->aNbr) { return p->aSym; }
  }
  return "";
}

void reformAtomicSymbol(string &s) {
  long ls = s.length(), p;
  for (p = 0; p < ls; p++) {
    if (p == 0) { s[p] = toupper(s[p]); }
    else { s[p] = tolower(s[p]); }
  }
}

/*----- tableOfMethods -------------------------------------------------*/

tableOfMethods::tableOfMethods() {
  methodInfo m;
  clear();
  m.method = "HF"; m.basis = "6-31G(D)"; m.scaleFactVib = 0.8929;
   m.scaleFactZPE = 0.9135; m.scaleFactGT = 1. / 1.12; push_back(m);
  m.method = "B3LYP"; m.basis = "6-31G(D)"; m.scaleFactVib = 0.9613;
   m.scaleFactZPE = 0.9804; m.scaleFactGT = 0.96; push_back(m);
  m.method = "B3LYP"; m.basis = "6-311G(D,P)"; m.scaleFactVib = 0.97;
   m.scaleFactZPE = 0.99; m.scaleFactGT = 0.99; push_back(m);
  m.method = "B3LYP"; m.basis = "CBSB7"; m.scaleFactVib = 0.97;
   m.scaleFactZPE = 0.99; m.scaleFactGT = 0.99; push_back(m);
}

int tableOfMethods::getIdx(string &rt) {
  for (int i = 0; i < size(); i++) {
    if ( (rt.find((*this)[i].method) != string::npos) &&
         (rt.find((*this)[i].basis) != string::npos) )
      { return i; }
  }
  return -1;
}

/*------------------------------------------------------------------------
     Generic Vector
------------------------------------------------------------------------*/

                                                  //----- constructors ---
gnVector::gnVector() { clear(); }
gnVector::gnVector(int siz) { init(siz); }

void gnVector::init(int siz) {                    //------------- init ---
  int i; double zero = 0.;
  clear(); for (i = 0; i < siz; i++) { push_back(zero); }
}

double gnVector::operator*(gnVector gv) {         //------ *(gnVector) ---
  // inner product
  int i, n = this->size();
  double prod = 0;
  if (gv.size() < n) { n = gv.size(); }
  for (i = 0; i < n; i++) { prod += (*this)[i] * gv[i]; }
  return prod;
}

/*------------------------------------------------------------------------
     Generic Matrix
------------------------------------------------------------------------*/

                                                  //----- constructors ---
gnMatrix::gnMatrix() { clear(); }
gnMatrix::gnMatrix(int rsiz, int csiz) { init(rsiz, csiz); }

void gnMatrix::init(int rsiz, int csiz) {         //------------- init ---
  int irow; gnVector zero(csiz);
  clear(); for (irow = 0; irow < rsiz; irow++) { push_back(zero); }
}

gnVector gnMatrix::operator*(gnVector gv) {       //------ *(gnVector) ---
  int rsiz = this->size(), csiz, irow;
  gnVector prod;
  if (rsiz == 0) { return prod; }
  csiz = (*this)[0].size(); if (gv.size() < csiz) { csiz = gv.size(); }
  if (csiz == 0) { return prod; }
  prod.init(rsiz);
  for (irow = 0; irow < rsiz; irow++) { prod[irow] = (*this)[irow] * gv; }
  return prod;
}

gnMatrix gnMatrix::operator~() {                  //-------------- ~() ---
  // transposition
  int rsiz = this->size(), csiz, irow, icol;
  gnMatrix trnp;
  if (rsiz == 0) { return trnp; }
  csiz = (*this)[0].size(); if (csiz == 0) { return trnp; }
  trnp.init(csiz, rsiz);
  for (irow = 0; irow < rsiz; irow++) {
    for (icol = 0; icol < csiz; icol++)
      { trnp[icol][irow] = (*this)[irow][icol]; }
  }
  return trnp;
}

gnMatrix gnMatrix::sqElement() {                  //-------- sqElement ---
  int rsiz = this->size(), csiz, irow, icol;
  gnMatrix sqm;
  if (rsiz == 0) { return sqm; }
  csiz = (*this)[0].size(); if (csiz == 0) { return sqm; }
  sqm.init(rsiz, csiz);
  for (irow = 0; irow < rsiz; irow++) {
    for (icol = 0; icol < csiz; icol++)
      { sqm[irow][icol] = (*this)[irow][icol] * (*this)[irow][icol]; }
  }
  return sqm;
}

/*------------------------------------------------------------------------
     Cartesian coordinate (also a vector)
------------------------------------------------------------------------*/

/*----- coordCart ------------------------------------------------------*/

coordCart::coordCart() {
  int i; double zero = 0.;
  clear(); for (i = 0; i < 3; i++) { push_back(zero); }
}

                                                  //------ x, y, and z ---
double &coordCart::x() { return (*this)[IX]; }
double &coordCart::y() { return (*this)[IY]; }
double &coordCart::z() { return (*this)[IZ]; }

void coordCart::fillZero() {                      //--------- fillZero ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] = 0.; }
}

double coordCart::vctLength()                     //-------- vctLength ---
  { return sqrt(this->innerProduct(*this)); }

void coordCart::normalize()                       //-------- normalize ---
  { divideMe(vctLength()); }

coordCart coordCart::invert() {                   //----------- invert ---
  int ic;
  for (ic = 0; ic < 3; ic++) { (*this)[ic] = -(*this)[ic]; }
  return *this;
}

coordCart coordCart::operator-() { return invert(); }

coordCart coordCart::add(coordCart crd) {         //-------------- add ---
  int ixyz;
  coordCart rs;
  for (ixyz = 0; ixyz < 3; ixyz++)
    { rs[ixyz] = (*this)[ixyz] + crd[ixyz]; }
  return rs;
}

coordCart coordCart::operator+(coordCart crd) { return add(crd); }

coordCart coordCart::addMe(coordCart crd) {       //------------ addMe ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] += crd[ixyz]; }
  return *this;
}

coordCart coordCart::operator+=(coordCart crd) { return addMe(crd); }

coordCart coordCart::subtract(coordCart crd) {    //--------- subtract ---
  int ixyz;
  coordCart rs;
  for (ixyz = 0; ixyz < 3; ixyz++)
    { rs[ixyz] = (*this)[ixyz] - crd[ixyz]; }
  return rs;
}

coordCart coordCart::operator-(coordCart crd) { return subtract(crd); }

coordCart coordCart::subtractMe(coordCart crd) {  //------- subtractMe ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] -= crd[ixyz]; }
  return *this;
}

coordCart coordCart::operator-=(coordCart crd) { return subtractMe(crd); }

coordCart coordCart::multiply(double fac) {       //--------- multiply ---
  int ixyz;
  coordCart rs;
  for (ixyz = 0; ixyz < 3; ixyz++) { rs[ixyz] = (*this)[ixyz] * fac; }
  return rs;
}

coordCart coordCart::operator*(double fac) { return multiply(fac); }

coordCart coordCart::multiplyMe(double fac) {     //------- multiplyMe ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] *= fac; }
  return *this;
}

coordCart coordCart::operator*=(double fac) { return multiplyMe(fac); }

coordCart coordCart::divideMe(double den) {       //--------- divideMe ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] /= den; }
  return *this;
}

coordCart coordCart::operator/=(double den) { return divideMe(den); }

double coordCart::innerProduct(coordCart crd) {   //----- innerProduct ---
  int ixyz;
  double rs = 0.;
  for (ixyz = 0; ixyz < 3; ixyz++) { rs += (*this)[ixyz] * crd[ixyz]; }
  return rs;
}

double coordCart::operator*(coordCart crd) { return innerProduct(crd); }

coordCart coordCart::outerProduct(coordCart crd) {  //--- outerProduct ---
  int ixyz, i1, i2;
  coordCart prod;
  for (ixyz = 0; ixyz < 3; ixyz++) {
    i1 = cycIncCrdIdx(ixyz); i2 = cycIncCrdIdx(i1);
    prod[ixyz] = (*this)[i1] * crd[i2] - crd[i1] * (*this)[i2];
  }
  return prod;
}

double coordCart::dirCosine(coordCart crd) {      //-------- dirCosine ---
  double dircos;
  coordCart nrmv1 = (*this), nrmv2 = crd;
  nrmv1.normalize(); nrmv2.normalize();
  dircos = nrmv1 * nrmv2;
  return dircos;
}

double coordCart::angleWith(coordCart crd) {      //-------- angleWith ---
  double angcos;
  angcos = dirCosine(crd);
  return acos(angcos) / M_PI * 180.;
}

int coordCart::minAbsIndex() {                    //------ minAbsIndex ---
  // return index of minimum absolute value component
  int ixyz, idxMin = 0;
  double minVal = 1.e+30;
  for (ixyz = 0; ixyz < 3; ixyz++) {
    if (fabs((*this)[ixyz]) < minVal)
      { minVal = fabs((*this)[ixyz]); idxMin = ixyz; }
  }
  return idxMin;
}

ostream &coordCart::print(ostream &os) {          //------------ print ---
  int ixyz;
  os << "(";
  for (ixyz = 0; ixyz < 3; ixyz++) { os << ' ' << (*this)[ixyz]; }
  os << " )\n";
  return os;
}

ostream &operator<<(ostream &os, coordCart &crd) { return crd.print(os); }

coordCart operator*(double fac, coordCart crd)
  { return crd.multiply(fac); }

coordCart vectProd(coordCart crdA, coordCart crdB)
  { return crdA.outerProduct(crdB); }

int cycIncCrdIdx(int idx) {                     //------- cycIncCrdIdx ---
  int incIdx;
  incIdx = idx + 1;
  if (incIdx > 2) { incIdx -= 3; }
  return incIdx;
}

int cycDecCrdIdx(int idx) {                     //------- cycDecCrdIdx ---
  int decIdx;
  decIdx = idx - 1;
  if (decIdx < 0) { decIdx += 3; }
  return decIdx;
}

coordCart vertPnt(coordCart &a, coordCart &b, coordCart &c) {
  /*  returns coordinate of point p on the line though a and b
      and nearest to point c, i.e., (a-b) is vertical to (p-c).   */
  coordCart v, p;
  double d;
  v = b - a; d = ((c - a) * v) / (v * v);
  p = a + d * v;
  return p;
}

double vertDist(coordCart &a, coordCart &b, coordCart &c) {
  /*  returns the length of (c-p) for p on the line though a and b
      and nearest to point c, i.e., (a-b) is vertical to (p-c).   */
  double rs;
  coordCart p;
  p = vertPnt(a, b, c);
  rs = (p - c).vctLength();
  return rs;
}

/*------------------------------------------------------------------------
     Cartesian axis
------------------------------------------------------------------------*/

                                                  //------ setByPoints ---
void axisCart::setByPoints(coordCart &pnt0, coordCart &pnt1) {
  /* set by two points */
  origin = pnt0; direct = pnt1 - pnt0; direct.normalize();
}

                                                  //----- setByOrgVect ---
void axisCart::setByOrgVect(coordCart &org, coordCart &vct) {
  /* set by origin and direction vector */
  origin = org; direct = vct; direct.normalize();
}

coordCart axisCart::vertPoint(coordCart &pntC) {  //-------- vertPoint ---
  /* returns point P on the axis nearest to point C,
     i.e., C-P is perpendicular to the axis.         */
  double length = (pntC - origin) * direct;
  return origin + length * direct;
}

double axisCart::distance(coordCart &pntC) {      //--------- distance ---
  /* distance of pntC from the axis */
  coordCart pntP = vertPoint(pntC);
  return (pntC - pntP).vctLength();
}

/*------------------------------------------------------------------------
     Cartesian matrix
------------------------------------------------------------------------*/

/*----- matrxCart ------------------------------------------------------*/

matrxCart::matrxCart() {
  int i;
  coordCart zvect;
  clear(); for (i = 0; i < 3; i++) { push_back(zvect); }
}

                                                  //------- X, Y and Z ---
coordCart &matrxCart::X() { return (*this)[IX]; }
coordCart &matrxCart::Y() { return (*this)[IY]; }
coordCart &matrxCart::Z() { return (*this)[IZ]; }
double &matrxCart::X(int idxc) { return (*this)[IX][idxc]; }
double &matrxCart::Y(int idxc) { return (*this)[IY][idxc]; }
double &matrxCart::Z(int idxc) { return (*this)[IZ][idxc]; }

coordCart matrxCart::multVect(coordCart crd) {    //--------- multVect ---
  int ir, ic;
  coordCart prd;
  for (ir = 0; ir < 3; ir++) {
    for (ic = 0; ic < 3; ic++) { prd[ir] += (*this)[ir][ic] * crd[ic]; }
  }
  return prd;
}

coordCart matrxCart::operator*(coordCart crd) { return multVect(crd); }

double matrxCart::determinant() {                 //------ determinant ---
  double det;
  det = (*this)[0][0] * (*this)[1][1] * (*this)[2][2]
      + (*this)[0][1] * (*this)[1][2] * (*this)[2][0]
      + (*this)[0][2] * (*this)[1][0] * (*this)[2][1]
      - (*this)[0][2] * (*this)[1][1] * (*this)[2][0]
      - (*this)[0][0] * (*this)[1][2] * (*this)[2][1]
      - (*this)[0][1] * (*this)[1][0] * (*this)[2][2];
  return det;
}

/*------------------------------------------------------------------------
     Polar coordinate
------------------------------------------------------------------------*/

coordPolar::coordPolar()                          //------- coordPolar ---
 { r = 0.; th = 0.; ph = 0.; }

void coordPolar::setFrom(coordCart cc) {          //---------- setFrom ---
  double lxy;
  r = cc.vctLength();
  lxy = sqrt(cc[IX] * cc[IX] + cc[IY] * cc[IY]);
  th = acos(cc[IX] / lxy);
  if (cc[IY] < 0.) { th = -th; }
  ph = asin(cc[IZ] / r);
}

double coordPolar::degTh()                        //------------ degTh ---
 { return th / M_PI * 180.; }

double coordPolar::degPh()                        //------------ degPh ---
 { return ph / M_PI * 180.; }

coordPolar coordPolar::subtract(coordPolar cp) {  //--------- subtract ---
  coordPolar rs;
  rs.r = r - cp.r;
  rs.th = th - cp.th;
  if (rs.th > M_PI) { rs.th -= 2. * M_PI; }
  else if (rs.th < -M_PI) { rs.th += 2. * M_PI; }
  rs.ph = ph - cp.ph;
  return rs;
}

coordPolar coordPolar::operator-(coordPolar cp) { return subtract(cp); }

/*------------------------------------------------------------------------
     Gaussian Job Informations
------------------------------------------------------------------------*/

jobInfos::jobInfos() { clear(); }                 //---- (constructor) ---

void jobInfos::clear() {                          //------------ clear ---
  version.erase(); revision.erase(); comment.erase();
  methodIdx = -1; method = "unknown"; basis = "unknown";
  scaleFactVib = 1.; scaleFactZPE = 1.; scaleFactGT = 1.;
  SCFenergy = "0.";
}

                                                  //---------- setFrom ---
int jobInfos::setFrom(string &gauv, string &route, string &comm,
 string &scfe) {
  long peq, pau;
  clear();
  version = gauv.substr(10, 2); revision = gauv.substr(15);
  methodIdx = gmth.getIdx(route);
  setMethodProp();
  comment = comm;
  SCFenergy = "";
  strAllToUpper(scfe);
  if ((peq = scfe.find('=')) != string::npos) {
    if ((pau = scfe.find("A.U.")) != string::npos) {
      SCFenergy = trim(scfe.substr(peq + 1, pau - peq - 1));
    }
  }
  return NORM;
}

int jobInfos::restoreFrom(istream &is) {          //------ restoreFrom ---
  clear();
  if (!getlGpoMatch(is, "jobInfos{")) { return ERRE; }
  if (!getlGpoStrv(is, version)) { return ERRE; }
  if (!getlGpoStrv(is, revision)) { return ERRE; }
  if (!getlGpoStrv(is, comment)) { return ERRE; }
  if (!getlGpoIntv(is, methodIdx)) { return ERRE; }
  if (!getlGpoDblv(is, scaleFactVib)) { return ERRE; }
  if (!getlGpoDblv(is, scaleFactZPE)) { return ERRE; }
  if (!getlGpoDblv(is, scaleFactGT)) { return ERRE; }
  if (!getlGpoStrv(is, SCFenergy)) { return ERRE; }
  if (!getlGpoMatch(is, "}jobInfos")) { return ERRE; }
  // - removed Jan. 18, 2008
  // - re-activated June 11, 2009
  setMethodProp();
  return NORM;
}

void jobInfos::setMethodProp() {                  //---- setMethodProp ---
  if (methodIdx >= 0) {
    method = gmth[methodIdx].method; basis = gmth[methodIdx].basis;
    scaleFactVib = gmth[methodIdx].scaleFactVib;
    scaleFactZPE = gmth[methodIdx].scaleFactZPE;
    scaleFactGT = gmth[methodIdx].scaleFactGT;
  } else {
    method = "unknown"; basis = "unknown";
    scaleFactVib = 1.; scaleFactZPE = 1.; scaleFactGT = 1.;
  }
}

ostream &jobInfos::print(ostream &os) {
  int prec;
  os << "jobInfos{\n";
  os << "  version " << version << endl;
  os << "  revision " << revision << endl;
  os << "  comment " << comment << endl;
  os << "  methodIdx " << methodIdx << "  ! "
     <<  method << "/" << basis << endl;
  prec = os.precision(); os.precision(12);
  os << "  scaleFactVib " << scaleFactVib << endl;
  os << "  scaleFactZPE " << scaleFactZPE << endl;
  os << "  scaleFactGT " << scaleFactGT << endl;
  os.precision(prec);
  os << "  SCFenergy " << SCFenergy << endl;
  os << "}jobInfos\n";
  return os;
}

ostream &operator<<(ostream &os, jobInfos &ji) { return ji.print(os); }

/*------------------------------------------------------------------------
     Base classes for geometry and frequency
------------------------------------------------------------------------*/

/*----- devnInfo: A set of atomic deviations (normal vibration) --------*/

devnInfo devnInfo::add(devnInfo di) {             //-------------- add ---
  int ic, nc = size(), nca = di.size();
  coordCart crd;
  devnInfo rs;
  if (nca < nc) { nc = nca; }
  rs.clear();
  for (ic = 0; ic < nc; ic++)
    { crd = (*this)[ic] + di[ic]; rs.push_back(crd); }
  return rs;
}

devnInfo devnInfo::operator+(devnInfo di) { return add(di); }

devnInfo devnInfo::addMe(devnInfo di) {           //------------ addMe ---
  int ic, nc = size(), nca = di.size();
  if (nca < nc) { nc = nca; }
  for (ic = 0; ic < nc; ic++) { (*this)[ic] += di[ic]; }
  return *this;
}

devnInfo devnInfo::operator+=(devnInfo di) { return addMe(di); }

devnInfo devnInfo::subtract(devnInfo di) {        //--------- subtract ---
  int ic, nc = size(), ncs = di.size();
  coordCart crd;
  devnInfo rs;
  if (ncs < nc) { nc = ncs; }
  rs.clear();
  for (ic = 0; ic < nc; ic++)
    { crd = (*this)[ic] - di[ic]; rs.push_back(crd); }
  return rs;
}

devnInfo devnInfo::operator-(devnInfo di) { return subtract(di); }

devnInfo devnInfo::subtractMe(coordCart crd) {    //-- subtractMe(crd) ---
  int ic, nc = size();
  for (ic = 0; ic < nc; ic++) { (*this)[ic] -= crd; }
  return *this;
}

devnInfo devnInfo::operator-=(coordCart crd) { return subtractMe(crd); }

devnInfo devnInfo::subtractMe(devnInfo di) {      //--- subtractMe(di) ---
  int ic, nc = size(), ncs = di.size();
  if (ncs < nc) { nc = ncs; }
  for (ic = 0; ic < nc; ic++) { (*this)[ic] -= di[ic]; }
  return *this;
}

devnInfo devnInfo::operator-=(devnInfo di) { return subtractMe(di); }

devnInfo devnInfo::multiply(double fac) {         //--------- multiply ---
  devnInfo rs;
  coordCart crd;
  devnInfo::iterator p;
  rs.clear();
  for (p = this->begin(); p != this->end(); p++) {
    crd = (*p) * fac;
    rs.push_back(crd);
  }
  return rs;
}

devnInfo devnInfo::operator*(double fac) { return multiply(fac); }

devnInfo devnInfo::divideMe(double den) {         //-------- deviedeMe ---
  int ic, nc = size();
  for (ic = 0; ic < nc; ic++) { (*this)[ic] /= den; }
  return *this;
}

double devnInfo::sumInnerProd(devnInfo di) {      //----- sumInnerProd ---
  int ic, ncrd = size(), ncdi = di.size();
  double sumip = 0.;
  if (ncdi < ncrd) { ncrd = ncdi; }
  for (ic = 0; ic < ncrd; ic++) { sumip += (*this)[ic] * di[ic]; }
  return sumip;
}

double devnInfo::operator*(devnInfo di) { return sumInnerProd(di); }

double devnInfo::amplitude()                      //-------- amplitude ---
  { return sqrt(this->sumInnerProd(*this)); }

devnInfo devnInfo::normalize()                    //-------- normalize ---
  { return divideMe(amplitude()); }

devnInfo operator*(double fac, devnInfo di) { return di.multiply(fac); }

/*----- atomCoord: Atomic coordinate -----------------------------------*/

atomCoord::atomCoord()
 { atomicNumber = 0; atomicType = 0; }

atomCoord atomCoord::setCoord(coordCart crd) {    //--------- setCoord ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] = crd[ixyz]; }
  return *this;
}

atomCoord atomCoord::operator=(coordCart crd) { return setCoord(crd); }

atomCoord atomCoord::add(coordCart crd) {         //-------------- add ---
  int ixyz;
  atomCoord rs = *this;
  for (ixyz = 0; ixyz < 3; ixyz++) { rs[ixyz] += crd[ixyz]; }
  return rs;
}

atomCoord atomCoord::operator+(coordCart crd) { return add(crd); }

atomCoord atomCoord::subtractMe(coordCart crd) {  //------- subtractMe ---
  int ixyz;
  for (ixyz = 0; ixyz < 3; ixyz++) { (*this)[ixyz] -= crd[ixyz]; }
  return *this;
}

atomCoord atomCoord::operator-=(coordCart crd) { return subtractMe(crd); }

/*------------------------------------------------------------------------
     Geometry
------------------------------------------------------------------------*/

/*----- geomInfo -------------------------------------------------------*/

int geomInfo::restoreFrom(istream &is) {          //------ restoreFrom ---
  int nAtms, iatm;
  atomCoord ac;
  clear();
  if (!getlGpoIntv(is, nAtms)) { return ERRE; }
  if (!getlGpoMatch(is, "geom{")) { return ERRE; }
  for (iatm = 0; iatm < nAtms; iatm++) {
    // --- changed from "getlGpoAtmCoord": June 13, 2009
    if (!getlAmbAtmCoord(is, ac)) { return ERRE; }
    push_back(ac);
  }
  if (!getlGpoMatch(is, "}geom")) { return ERRE; }
  setPrcAxesMomInert();
  return NORM;
}

void geomInfo::outFWformat(ostream &os) {         //------ outFWformat ---
  int iatm, ixyz;
  os << "                   Input orientation:\n\n\n\n\n";
  for (iatm = 0; iatm < size(); iatm++) {
    os << ' ';
    os.width(4); os.setf(ios::right); os << iatm + 1;
    os << "       ";
    os.width(4); os.setf(ios::right); os << (*this)[iatm].atomicNumber;
    os << "       ";
    for (ixyz = 0; ixyz < 3; ixyz++) {
      os.precision(6); os.setf(ios::fixed | ios::right); os.width(12);
      os << (*this)[iatm][ixyz];
    }
    os << endl;
  }
  os << endl;
}

void geomInfo::setMolW() {                        //---------- setMolW ---
  int ia, na = size();
  molW = 0.;
  for (ia = 0; ia < na; ia++)
    { molW += elm.getAtmcWeight((*this)[ia].atomicNumber); }
}

void geomInfo::setCentOfMass() {                  //---- setCentOfMass ---
  int iatm, ixyz;
  double atmcWgt;
  setMolW();
  com.fillZero();
  for (iatm = 0; iatm < size(); iatm++) {
    atmcWgt = elm.getAtmcWeight((*this)[iatm].atomicNumber);
    for (ixyz = 0; ixyz < 3; ixyz++)
      { com[ixyz] += atmcWgt * (*this)[iatm][ixyz]; }
  }
  for (ixyz = 0; ixyz < 3; ixyz++) { com[ixyz] /= molW; }
}

void geomInfo::setPrcAxesMomInert() {        //---- setPrcAxesMomInert ---
  const int lw = 200;
  int irow, icol, info;
  double matI[3][3], w[lw];
  matrxCart tnsI = getTensorInert();

  for (irow = 0; irow < 3; irow++) {
    matI[irow][irow] = tnsI[irow][irow];
    for (icol = irow + 1; icol < 3; icol++) {
      matI[icol][irow] = tnsI[irow][icol];
    }
  }
  info = lapack_dsyev('V', 'U', 3, (double **)matI, 3, momInert, w, lw);
  for (icol = 0; icol < 3; icol++) {
    for (irow = 0; irow < 3; irow++)
      { prcAxes[icol][irow] = matI[icol][irow]; }
  }
  // --- align principal axes right-handed
  if (prcAxes.determinant() < 0.) { prcAxes.Z() = -prcAxes.Z(); }
}

matrxCart geomInfo::getTensorInert(){             //--- getTensorInert ---
  int ia, na = size(), ic, ic_pls, ic_mns;
  double m;
  coordCart smC2, smXt, crd;
  matrxCart tnI;

  setCentOfMass();
  for (ia = 0; ia < na; ia++) {
    m = elm.getAtmcWeight((*this)[ia].atomicNumber);
    crd = (*this)[ia] - com;
    for (ic = 0; ic < 3; ic++) {
      smC2[ic] += m * crd[ic] * crd[ic];
      ic_pls = cycIncCrdIdx(ic); ic_mns = cycDecCrdIdx(ic);
      smXt[ic] += m * crd[ic_pls] * crd[ic_mns];
    }
  }
  for (ic = 0; ic < 3; ic++) {
    ic_pls = cycIncCrdIdx(ic); ic_mns = cycDecCrdIdx(ic);
    tnI[ic][ic] = smC2[ic_pls] + smC2[ic_mns];
    tnI[ic_pls][ic_mns] = -smXt[ic];
    tnI[ic_mns][ic_pls] = -smXt[ic];
  }
  return tnI;
}

geomInfo geomInfo::add(devnInfo di) {             //-------------- add ---
  int iatm;
  geomInfo rs;
  atomCoord ac;
  for (iatm = 0; iatm < size(); iatm++)
    { ac = (*this)[iatm] + di[iatm]; rs.push_back(ac); }
  return rs;
}

geomInfo geomInfo::operator+(devnInfo di) { return add(di); }

geomInfo geomInfo::subtractMe(coordCart crd) {    //------- subtractMe ---
  int ia, na = size();
  for (ia = 0; ia < na; ia++) { (*this)[ia] -= crd; }
  return *this;
}

geomInfo geomInfo::operator-=(coordCart crd) { return subtractMe(crd); }

void geomInfo::transformMe(matrxCart mc) {        //------ transformMe ---
  int ia, na = size();
  for (ia = 0; ia < na; ia++) { (*this)[ia] = mc * (*this)[ia]; }
}

void geomInfo::beCanonical() {                    //------ beCanonical ---
  setPrcAxesMomInert();
  (*this) -= com;
  transformMe(prcAxes);
  setPrcAxesMomInert();
}

coordCart geomInfo::getVelCOM(devnInfo &di) {     //-------- getVelCOM ---
  //--- get COM velocity vector for devnInfo di
  int ia, na = size();
  double summ = 0., m;
  coordCart vCOM;
  for (ia = 0; ia < na; ia++) {
    m = elm.getAtmcWeight((*this)[ia].atomicNumber);
    summ += m; vCOM +=  m * di[ia];
  }
  vCOM /= summ;
  return vCOM;
}

                                                  //----- getAngVPrcAx ---
double geomInfo::getAngVPrcAx(devnInfo &di, int iax) {
  //--- get angular velocity around the principal axes iax for devInfo di
  int ia, na = size();
  double angvel = 0., m;
  devnInfo unitRot = getUnitRotPrcAx(iax);
  for (ia = 0; ia < na; ia++) {
    m = elm.getAtmcWeight((*this)[ia].atomicNumber);
    angvel += m * (unitRot[ia] * di[ia]);
  }
  angvel /= (*this).momInert[iax];
  return angvel;
}

devnInfo geomInfo::getUnitRotPrcAx(int iax) {     //-- getUnitRotPrcAx ---
  //--- get unit rotation deviation vector around the principal axis iax
  int ia, na = size();
  double r;
  coordCart apnt = com + prcAxes[iax], vpnt, offax, dirrot;
  devnInfo unitRot;
  unitRot.clear();
  for (ia = 0; ia < na; ia++) {
    vpnt = vertPnt(com, apnt, (*this)[ia]);
    offax = (*this)[ia] - vpnt;
    if ((r = offax.vctLength()) < GAUCPRC) { dirrot.fillZero(); }
    else {
      dirrot = vectProd(prcAxes[iax], offax); dirrot.normalize();
      dirrot *= r;
    }
    unitRot.push_back(dirrot);
  }
  return unitRot;
}

ostream &geomInfo::print(ostream &os) {                   // --- print ---
  int prec, iatm, nAtms = size(), ixyz;
  prec = os.precision(); os.setf(ios::fixed); os.precision(6);
  os << " nAtms " << nAtms << endl;
  os << " molW " << molW << endl;
  os << " COM " << com;
  os << " principal Axes:\n";
  for (ixyz = 0; ixyz < 3; ixyz++) {
    os << "  axis-" << (ixyz + 1) << prcAxes[ixyz];
  }
  os << " momI";
  for (ixyz = 0; ixyz < 3; ixyz++) { os << ' ' << momInert[ixyz]; }
  os << endl;
  os << " geom{\n";
  for (iatm = 0; iatm < nAtms; iatm++) {
    os << "   " << (*this)[iatm].atomicNumber << ' '
       << (*this)[iatm].atomicType;
    for (ixyz = 0; ixyz < 3; ixyz++) { os << ' ' << (*this)[iatm][ixyz]; }
    os << endl;
  }
  os.unsetf(ios::fixed); os.precision(prec);
  os << " }geom\n";
  return os;
}

ostream &operator<<(ostream &os, geomInfo &gi) { return gi.print(os); }

/*----- molecGeom ------------------------------------------------------*/

molecGeom::molecGeom() {                          // --- (constructor) ---
  int i;
  nAtms = 0; gElec = 1; nVibm = 0; isLinear = false;
  rotSymNbr = 1.; numIsomers = 1.; molW = 0.;
  for (i = 0; i < 3; i++) { rotcGHz[i] = 0.; }
}

                                                  //---------- setFrom ---
int molecGeom::setFrom(vector<string> &gm, string &estate,
 string &chgmul) {
  int il, cl, ia, ic, anum, natm_one = 1;
  int irot, geoTop, geoBot, exType = false, colSh = 0, pg1n;
  long p, pdash;
  string s;
  char c, pg0, pg1, pg2;
  atomCoord crd;

  if ((p = chgmul.find("Multiplicity")) != string::npos)
    { gElec = atoi(trim(chgmul.substr(p + 14)).c_str()); }
  else { gElec = 1; }
  for (il = 0, cl = 0; il < gm.size(); il++) {
    if (gm[il].substr(1,10) == "----------") {
      cl++;
      if (cl == 2) { geoTop = il + 1; }
      else if (cl == 3) { geoBot = il - 1; break; }
    }
  }
  nAtms = geoBot - geoTop + 1;
  if (nAtms < 1) { return ERRE; }

  s = gm[0].substr(18);
  if ((p = s.find('[')) == string::npos) {
    pntGrp = s;
  } else {
    pntGrp = s.substr(0, p);
  }
  if ((pntGrp == "D*H") || (pntGrp == "C*V")) {
    isLinear = true; nVibm = 3 * nAtms - 5;
  } else {
    isLinear = false; nVibm = 3 * nAtms - 6;
  }

  chemform.clear();
  geom.clear();
  if (gm[geoTop - 2].find("Type") != string::npos)
    { exType = true; colSh = 11; }
  for (ia = 0; ia < nAtms; ia++) {
    s = gm[geoTop + ia];
    anum = atoi(s.substr(12, 4).c_str());
    if (chemform.find(anum) != chemform.end()) { chemform[anum]++; }
    else { chemform.insert(make_pair(anum, natm_one)); }
    crd.atomicNumber = anum;
    if (exType) { crd.atomicType = atoi(s.substr(19, 11).c_str()); }
    else { crd.atomicType = 0; }
    for (ic = 0; ic < 3; ic++)
      { crd[ic] = atof(s.substr(colSh + 24 + 12 * ic, 11).c_str()); }
    geom.push_back(crd);
  }

  geom.setPrcAxesMomInert();
  molW = geom.molW;
  s = gm[gm.size() - 1];
  for (irot = 0; irot < 3; irot++) {
    p = irot * 15 + 29;
    rotcGHz[irot] = atof(s.substr(p, 14).c_str());
  }
  pg0 = pntGrp[0]; pg1 = pntGrp[1];
  if (isdigit(pg1)) { pg1n = (int)pg1 - (int)'0'; }
  else { pg1n = 1; }
  if (pntGrp.length() > 2) { pg2 = pntGrp[2]; }
  else { pg2 = ' '; }
  switch (pg0) {
   case 'C': rotSymNbr = 1. * pg1n; break;
   case 'D': rotSymNbr = 2. * pg1n; break;
   case 'S': if (pg1n == 6) { rotSymNbr = 3.; } else { rotSymNbr = 1.; }
    break;
   case 'T': rotSymNbr = 12.; break;
   case 'O': rotSymNbr = 24.; break;
   default: rotSymNbr = 1.; break;
  }
  numIsomers = 1.;
  if (((pg0 == 'C') || (pg0 == 'D')) && (pg2 == ' ')) {
    if (pntGrp.substr(0, 2) != "CS") { numIsomers = 2.; }
  }
  if ((p = estate.find("electronic state is")) != string::npos) {
    elState = estate.substr(p + 20, estate.length() - p - 21);
    pdash = elState.find('-');
    if (isLinear) {
      if (elState.substr(pdash + 1, 2) != "SG") { gElec *= 2; }
    } else {
      if (elState[pdash + 1] == 'E') { gElec *= 2; }
    }
  } else { elState = ""; }
  return NORM;
}

void molecGeom::clear() {                         //------------ clear ---
  chemform.clear(); geom.clear(); pntGrp.erase();
}

int molecGeom::restoreFrom(istream &is) {         //------ restoreFrom ---
  int iatm;
  atomCoord ac;
  clear();
  if (!getlGpoMatch(is, "molecGeom{")) { return ERRE; }
  if (!getlGpoIntv(is, nAtms)) { return ERRE; }
  if (!getlGpoIntv(is, gElec)) { return ERRE; }
  if (!getlGpoIntv(is, nVibm)) { return ERRE; }
  if (!getlGpoBlnv(is, isLinear)) { return ERRE; }
  if (!getlGpoStrv(is, pntGrp)) { return ERRE; }
  if (!getlGpoStrv(is, elState)) { return ERRE; }
  if (!getlGpoDblv(is, rotSymNbr)) { return ERRE; }
  if (!getlGpoDblv(is, numIsomers)) { return ERRE; }
  if (!getlGpoDblv(is, molW)) { return ERRE; }
  if (!getlGpoDblvMlt(is, rotcGHz, 3)) { return ERRE; }
  if (!getlGpoChmFrm(is, chemform)) { return ERRE; }
  if (!getlGpoMatch(is, "geom{")) { return ERRE; }
  geom.clear();
  for (iatm = 0; iatm < nAtms; iatm++) {
    if (!getlGpoAtmCoord(is, ac)) { return ERRE; }
    geom.push_back(ac);
  }
  if (!getlGpoMatch(is, "}geom")) { return ERRE; }
  if (!getlGpoMatch(is, "}molecGeom")) { return ERRE; }
  geom.setPrcAxesMomInert();
  return NORM;
}

ostream &molecGeom::print(ostream &os) {          //------------ print ---
  int irot, iatm, ixyz, prec;
  map<int, int>::iterator p;
  os << "molecGeom{\n";
  os << "  nAtms " << nAtms << endl;
  os << "  gElec " << gElec << endl;
  os << "  nVibm " << nVibm << endl;
  os << "  isLinear " << strGpoBlnv(isLinear) << endl;
  os << "  pntGrp " << pntGrp << endl;
  os << "  elState " << elState << endl;
  os << "  rotSymNbr " << rotSymNbr << endl;
  os << "  numIsomers " << numIsomers << endl;
  prec = os.precision(); os.setf(ios::fixed);
  os.precision(6); os << "  molW " << molW << endl;
  os.precision(7); os << "  rotcGHz";
  for (irot = 0; irot < 3; irot++) { os << ' ' << rotcGHz[irot]; }
  os << endl;
  os << "  chemform";
  for (p = chemform.begin(); p != chemform.end(); p++)
    { os << ' ' << elm.getAtmcSym(p->first) << ' ' << p->second; }
  os << endl;
  os.precision(6); os << "  geom{\n";
  for (iatm = 0; iatm < nAtms; iatm++) {
    os << "    " << geom[iatm].atomicNumber << ' '
       << geom[iatm].atomicType;
    for (ixyz = 0; ixyz < 3; ixyz++) { os << ' ' << geom[iatm][ixyz]; }
    os << endl;
  }
  os.unsetf(ios::fixed); os.precision(prec);
  os << "  }geom\n";
  os << "}molecGeom\n";
  return os;
}

ostream &operator<<(ostream &os, molecGeom &mg) { return mg.print(os); }

/*------------------------------------------------------------------------
     Frequencies
------------------------------------------------------------------------*/

/*----- freqTable ------------------------------------------------------*/

freqTable::freqTable()                            // --- (constructor) ---
  { clear(); zpe_Unscaled = 0.; zpe = 0.; zpe_GT = 0.; }

                                                  //---------- setFrom ---
int freqTable::setFrom(vector<string> &fr, jobInfos &ji, molecGeom &mg) {
  int iblk = 0, topL, idxStt, idxEnd, iEnd, i, idxRead, pos, ip;
  int nParam = 6, ia, ixyz;
  string s;
  double pars[3][7], sumf;
  coordCart nrmm;
  freqInfo fi[3];

  clear();
  for (i = 6; i < 10; i++) {
    s = fr[i];
    if (s.substr(1, 7) == "Atom AN") { nParam = i - 2; break; }
  }
  do {
    idxStt = iblk * 3; idxEnd = idxStt + 2; iEnd = 2;
    if (idxEnd > (mg.nVibm - 1))
      { idxEnd = (mg.nVibm - 1); iEnd = idxEnd - idxStt; }
    topL = iblk * (mg.nAtms + nParam + 3);
    s = fr[topL];
    for (i = 0; i <= iEnd; i++) {
      pos = i * 23 + 18; idxRead = atoi(s.substr(pos, 4).c_str());
      if (idxRead != (idxStt + i + 1)) { return ERRE; }
    }
    s = fr[topL + 1];
    for (i = 0; i <= iEnd; i++)
      { pos = i * 23 + 20; fi[i].sym = trim(s.substr(pos,  6)); }
    for (ip = 0; ip < nParam; ip++) {
      s = fr[topL + ip + 2];
      for (i = 0; i <= iEnd; i++) {
        pos = i * 23 + 15; pars[i][ip] = atof(s.substr(pos, 11).c_str());
      }
    }
    for (i = 0; i <= iEnd; i++) {
      fi[i].freq = pars[i][0]; fi[i].rmass = pars[i][1];
      fi[i].fconst = pars[i][2]; fi[i].irint = pars[i][3];
      if (nParam >= 6) {
        fi[i].ramact = pars[i][4]; fi[i].depol = pars[i][5];
      } else {
        fi[i].ramact = 0.; fi[i].depol = 0.;
      }
    }
    for (i = 0; i <= iEnd; i++) { fi[i].nrm.clear(); }
    for (ia = 0; ia < mg.nAtms; ia++) {
      s = fr[topL + nParam + ia + 3];
      for (i = 0; i <= iEnd; i++) {
        for (ixyz = 0; ixyz < 3; ixyz++) {
          pos = i * 23 + ixyz * 7 + 11;
          nrmm[ixyz] = atof(s.substr(pos, 6).c_str());
        }
        fi[i].nrm.push_back(nrmm);
      }
    }
    for (i = 0; i <= iEnd; i++) {
      push_back(fi[i]);
    }
    iblk++;
  } while ((iblk * 3) < mg.nVibm);
  sumf = 0.;
  for (iterator pf = begin(); pf != end(); pf++) {
    if (pf->freq > 0.) { sumf += pf->freq; }
  }
  zpe_Unscaled = sumf / 2.;
  zpe = sumf * ji.scaleFactZPE / 2.; zpe_GT = sumf * ji.scaleFactGT / 2.;
  return NORM;
}

int freqTable::restoreFrom(istream &is) {         //------ restoreFrom ---
  int nVibm, ifrq, nAtms, iatm;
  freqInfo fi;
  coordCart crd;
  clear();
  if (!getlGpoMatch(is, "freqTable{")) { return ERRE; }
  if (!getlGpoDblv(is, zpe_Unscaled)) { return ERRE; }
  if (!getlGpoDblv(is, zpe)) { return ERRE; }
  if (!getlGpoDblv(is, zpe_GT)) { return ERRE; }
  if (!getlGpoIntv(is, nVibm)) { return ERRE; }
  for (ifrq = 0; ifrq < nVibm; ifrq++) {
    if (!getlGpofrqInfo(is, fi)) { return ERRE; }
    if (!getlGpoIntv(is, nAtms)) { return ERRE; }
    fi.nrm.clear();
    for (iatm = 0; iatm < nAtms; iatm++) {
      if (!getlGpoCrdCart(is, crd)) { return ERRE; }
      fi.nrm.push_back(crd);
    }
    if (!getlGpoMatch(is, "}nrm")) { return ERRE; }
    push_back(fi);
  }
  if (!getlGpoMatch(is, "}frequencies")) { return ERRE; }
  if (!getlGpoMatch(is, "}freqTable")) { return ERRE; }
  return NORM;
}

ostream &freqTable::print(ostream &os) {          //------------ print ---
  int ifrq, iatm, ixyz, prec;
  os << "freqTable{\n";
  prec = os.precision(); os.setf(ios::fixed);
  os << "  zpe_Unscaled ";
   os.precision(5); os << zpe_Unscaled << "  ! = ";
   os.precision(10); os << (zpe_Unscaled / WAVNHRT) << " hartree" << endl;
  os << "  zpe ";
   os.precision(5); os << zpe << "  ! = ";
   os.precision(10); os << (zpe / WAVNHRT) << " hartree" << endl;
  os << "  zpe_GT ";
   os.precision(5); os << zpe_GT << "  ! = ";
   os.precision(10); os << (zpe_GT / WAVNHRT) << " hartree" << endl;
  os << "  frequencies{ " << size() << endl;
  os.precision(4);
  for (ifrq = 0; ifrq < size(); ifrq++) {
    os << "    " << (*this)[ifrq].sym << ' ' << (*this)[ifrq].freq
       << ' ' << (*this)[ifrq].rmass << ' ' << (*this)[ifrq].fconst
       << ' ' << (*this)[ifrq].irint << ' ' << (*this)[ifrq].ramact
       << ' ' << (*this)[ifrq].depol << endl;
    os << "    nrm{ " << (*this)[ifrq].nrm.size() << endl;
    os.precision(2);
    for (iatm = 0; iatm < (*this)[ifrq].nrm.size(); iatm++) {
      os << "     ";
      for (ixyz = 0; ixyz < 3; ixyz++)
        { os << ' ' << (*this)[ifrq].nrm[iatm][ixyz]; }
      os << endl;
    }
    os.precision(4);
    os << "    }nrm\n";
  }
  os.unsetf(ios::fixed); os.precision(prec);
  os << "  }frequencies\n";
  os << "}freqTable\n";
  return os;
}

int freqTable::countImag() {                      //-------- countImag ---
  int numImg = 0, n, i;
  n = size();
  for (i = 0; i < n; i++) { if ((*this)[i].freq < 0.) { numImg++; } }
  return numImg;
}

                                                  //----- getMostAlike ---
vector<pair<int, double> > freqTable::getMostAlike(devnInfo &di) {
  //--- get modes most resembling di
  int ifq, nfq = size(), cnt;
  double fsim;
  vector<pair<int, double> > res;
  map<double, int> wrk;
  map<double, int>::iterator p;
  wrk.clear();
  for (ifq = 0; ifq < nfq; ifq++) {
    fsim = -fabs((*this)[ifq].nrm * di);
    wrk.insert(make_pair(fsim, ifq));
  }
  res.clear();
  for (p = wrk.begin(), cnt = 0; p != wrk.end(); p++) {
    if (cnt < 3) {
      res.push_back(make_pair(p->second, -(p->first))); cnt++;
    } else {
      if (p->first < -0.5)
        { res.push_back(make_pair(p->second, -(p->first))); cnt++; }
      else { break; }
    }
  }
  return res;
}

vibSimList freqTable::findSimil(devnInfo &di) {   //-------- findSimil ---
  int ifq, nfq = size(), cnt;
  double fsim;
  vibSimil vibs;
  vibSimList res;
  map<double, int> wrk;
  map<double, int>::iterator p;
  devnInfo ref, tgt;

  wrk.clear(); ref = di; ref.normalize();
  for (ifq = 0; ifq < nfq; ifq++) {
    tgt = (*this)[ifq].nrm; tgt.normalize();
    fsim = -fabs(tgt * ref);
    wrk.insert(make_pair(fsim, ifq));
  }
  res.clear();
  for (p = wrk.begin(), cnt = 0; p != wrk.end(); p++) {
    vibs.idx = p->second; vibs.simil = -(p->first);
    vibs.freq = (*this)[vibs.idx].freq;
    if (cnt < 3) { res.push_back(vibs); cnt++; }
    else if (p->first < -0.5) { res.push_back(vibs); cnt++; }
    else { break; }
  }
  return res;
}

ostream &operator<<(ostream &os, freqTable &ft) { return ft.print(os); }

/*------------------------------------------------------------------------
     Molecular moiety (for internal rotor calculations)
------------------------------------------------------------------------*/

                                                  //------- setTopProp ---
void molMoiety::setTopProp(atomCoord &cPiv, molecGeom &mg) {
  // --- top only
  setAppTopProp(cPiv);
  // --- asymmetric top
  defineTopAxes(cPiv);
  setAsymTopProp(cPiv, mg);
}

                                                  //------- setRotDevn ---
void molMoiety::setRotDevn(atomCoord &cPiv, devnInfo &rotDI) {
  int ia, na = geom.size();
  double r;
  coordCart dev, axRot, vpnt, offAx;
  dev.fillZero(); rotDI[lstAtms[0]] = dev;
  axRot = geom[0] - cPiv;
  for (ia = 1; ia < na; ia++) {
    vpnt = vertPnt(cPiv, geom[0], geom[ia]);
    offAx = geom[ia] - vpnt;
    if ((r = offAx.vctLength()) < GAUCPRC) { dev.fillZero(); }
    else { dev = vectProd(axRot, offAx); dev.normalize(); dev *= r; }
    rotDI[lstAtms[ia]] = dev;
  }
  rotDI.normalize();
}

void molMoiety::setAppTopProp(atomCoord &cPiv) {  //---- setAppTopProp ---
  int ia;
  double mass, r;
  coordCart crd;
  appI = 0.;
  for (ia = 1; ia < geom.size(); ia++) {
    crd = geom[ia];
    mass = elm.getAtmcWeight(geom[ia].atomicNumber);
    r = vertDist(cPiv, geom[0], crd);
    appI += mass * r * r;
  }
  appB = BIWAA2 / appI;
}

void molMoiety::defineTopAxes(atomCoord &cPiv) {  //---- defineTopAxes ---
  int iatm, ixyz, idxMin;
  double sumMass = 0., atmcWgt;
  coordCart wp, wpv;
  // - Top center-of-mass (excluding pivot)
  topCOM.fillZero();
  for (iatm = 1; iatm < geom.size(); iatm++) {
    atmcWgt = elm.getAtmcWeight(geom[iatm].atomicNumber);
    sumMass += atmcWgt;
    for (ixyz = 0; ixyz < 3; ixyz++)
      { topCOM[ixyz] += atmcWgt * geom[iatm][ixyz]; }
  }
  for (ixyz = 0; ixyz < 3; ixyz++) { topCOM[ixyz] /= sumMass; }
  // - asymmetric top coordinates (as in Pitzer-II)
  topAxes[IZ] = geom[0] - cPiv; topAxes[IZ].normalize();
  topOrigin = vertPnt(cPiv, geom[0], topCOM);
  topAxes[IX] = topCOM - topOrigin;
  if (topAxes[IX].vctLength() < (GAUCPRC * 2.)) {
    idxMin = topAxes[IZ].minAbsIndex();
    wp = topOrigin + topAxes[IZ]; wp[idxMin] += 1.;
    wpv = vertPnt(cPiv, geom[0], wp);
    topAxes[IX] = wp - wpv;
  }
  topAxes[IX].normalize();
  topAxes[IY] = vectProd(topAxes[IX], topAxes[IZ]);
  topAxes[IY].normalize();
}

                                                  //--- setAsymTopProp ---
void molMoiety::setAsymTopProp(atomCoord &cPiv, molecGeom &mg) {
  int ir, ic, iatm, irinc, irdec;
  double detAlp, atmcWgt, lambda = 0.;
  double Am, Bm, Cm, Um;
  coordCart topCrd, rvec, betvec;
  matrxCart alpmat, trmat, trprcm;
  // --- alpha matrix
  do {
    for (ir = 0; ir < 3; ir++) {
      for (ic = 0; ic < 3; ic++)
        { alpmat[ir][ic] = mg.geom.prcAxes[ir].dirCosine(topAxes[ic]); }
    }
    detAlp = alpmat.determinant();
    if (detAlp < 0.) {
      topAxes[IY] = topAxes[IY] * -1.;
    }
  } while (detAlp < 0.);
  // --- Am, Bm, Cm, and Um
  Am = 0.; Bm = 0.; Cm = 0.; Um = 0.;
  for (ir = 0; ir < 3; ir++) { trmat[ir] = topAxes[ir]; }
  for (iatm = 1; iatm < geom.size(); iatm++) {
    atmcWgt = elm.getAtmcWeight(geom[iatm].atomicNumber);
    topCrd = trmat * (geom[iatm] - topOrigin);
    Am += atmcWgt * (topCrd[IX] * topCrd[IX] + topCrd[IY] * topCrd[IY]);
    Bm += atmcWgt * topCrd[IX] * topCrd[IZ];
    Cm += atmcWgt * topCrd[IY] * topCrd[IZ];
    Um += atmcWgt * topCrd[IX];
  }
  // --- r (vector from molecular COM to top origin)
  for (ir = 0; ir < 3; ir++) { trprcm[ir] = mg.geom.prcAxes[ir]; }
  rvec = trprcm * (topOrigin - mg.geom.com);
  // --- beta vector
  for (ir = 0; ir < 3; ir++) {
    irinc = cycIncCrdIdx(ir); irdec = cycDecCrdIdx(ir);
    betvec[ir] = alpmat[ir][IZ]*Am - alpmat[ir][IX]*Bm - alpmat[ir][IY]*Cm
     + Um*(alpmat[irdec][IY]*rvec[irinc] - alpmat[irinc][IY]*rvec[irdec]);
  }
  // --- lambda
  for (ir = 0; ir < 3; ir++) {
    lambda += (pow(alpmat[ir][IY]*Um, 2) / mg.molW
              + pow(betvec[ir], 2) / mg.geom.momInert[ir]);
  }
  // --- reduced I and B for asymmetric top
  asmRedI = Am - lambda;
  asmRedB = BIWAA2 / asmRedI;
}

/*------------------------------------------------------------------------
     Molecule consisting of two molMoiety's (internal rotor)
------------------------------------------------------------------------*/

                                                  //---------- setList ---
int molByMoieties::setList(vector<string> &moin, molecGeom &mg) {
  int im, ia, idxAtm, lastWildC = false;
  vector<string> asLst;
  set<int> usedIdx;
  molMoiety mmo;
  coordCart crd0;

  clear();
  // --- parse input string
  for (im = 0; im < 2; im++) {
    asLst = splitTokens(moin[im], "-");
    mmo.lstAtms.clear(); mmo.geom.clear();
    for (ia = 0; ia < asLst.size(); ia++) {
      if (asLst[ia] == "@") {
        if (im == 0) { return ERRE; }
        else if (ia == 0) { return ERRE; }
        else { lastWildC = true; break; }
      } else { idxAtm = atoi(asLst[ia].c_str()) - 1; }
      if ((idxAtm < 0) || (idxAtm >= mg.nAtms)) { return ERRE; }
      if (usedIdx.find(idxAtm) != usedIdx.end()) { return ERRE; }
      mmo.lstAtms.push_back(idxAtm);
      mmo.geom.push_back(mg.geom[idxAtm]);
      usedIdx.insert(idxAtm);
    }
    push_back(mmo);
  }
  // --- expand '@'
  if (lastWildC) {
    for (idxAtm = 0; idxAtm < mg.nAtms; idxAtm++) {
      if (usedIdx.find(idxAtm) == usedIdx.end()) {
        (*this)[1].lstAtms.push_back(idxAtm);
        (*this)[1].geom.push_back(mg.geom[idxAtm]);
        usedIdx.insert(idxAtm);
      }
    }
  }
  if (usedIdx.size() == mg.nAtms) { isWholeMolec = true; }
  else { isWholeMolec = false; }
  // --- initialize intRotDevn
  crd0.fillZero(); intRotDevn.clear(); intRotDevn.resize(mg.nAtms, crd0);
  return NORM;
}

void molByMoieties::setProp(molecGeom &mg) {      //---------- setProp ---
  int ipa, im, imc;
  double fracI = 0., lambda;
  molMoiety *mm, *mmc;
  coordCart axPivPiv, onAxPnt, dirvct[2];

  axPivPiv = (*this)[0].geom[0] - (*this)[1].geom[0];
  for (ipa = 0; ipa < 3; ipa++) {
    lambda = axPivPiv.dirCosine(mg.geom.prcAxes[ipa]);
    fracI += lambda * lambda / mg.geom.momInert[ipa];
  }
  for (im = 0; im < 2; im++) {
    imc = (im + 1) % 2; mm = &(*this)[im]; mmc = &(*this)[imc];
    mm->setTopProp(mmc->geom[0], mg);
    mm->symRedI = mm->appI * (1. - mm->appI * fracI);
    mm->symRedB = BIWAA2 / mm->symRedI;
    onAxPnt = vertPnt(mmc->geom[0], mm->geom[0], mm->topCOM);
    dirvct[im] = mm->topCOM - onAxPnt;
  }
  dihedAngle = 0.;
  for (im = 0; im < 2; im++)
    { if (dirvct[im].vctLength() < GAUCPRC) { return; } }
  dihedAngle = dirvct[1].angleWith(dirvct[0]);
  if (axPivPiv * dirvct[1].outerProduct(dirvct[0]) < 0.)
    { dihedAngle *= (-1.); }
}

ostream &molByMoieties::print(ostream &os) {      //------------ print ---
  int im, ia;
  string whmol = "";
  molMoiety *mm[2] = {&(*this)[0], &(*this)[1]};
  for (im = 0; im < 2; im++) {
    os << "moiety-" << im + 1 << ": " << mm[im]->lstAtms[0] + 1;
    for (ia = 1; ia < mm[im]->lstAtms.size(); ia++)
      { os << "-" << mm[im]->lstAtms[ia] + 1; }
    os << endl;
  }
  if (!isWholeMolec) { whmol = " not"; }
  os << "two moieties are" << whmol << " whole molecule.\n";
  os.setf(ios::fixed); os.precision(4);
  os << "dihedral angle between moieties 1 and 2: " << dihedAngle << endl;
  for (im = 0; im < 2; im++) {
    os << "[moiety-" << im + 1 << "]\n";
    os.precision(8);
    os << "    appI[amuA2]: " << mm[im]->appI << endl;
    os << "    appB[cm-1]: " << mm[im]->appB << endl;
    os << " --- symmetric top ---\n";
    os << " symRedI[amuA2]: " << mm[im]->symRedI << endl;
    os << " symRedB[cm-1]: " << mm[im]->symRedB << endl;
    os << " --- asymmetric top ---\n";
    os << " asmRedI[amuA2]: " << mm[im]->asmRedI << endl;
    os << " asmRedB[cm-1]: " << mm[im]->asmRedB << endl;
  }
  return os;
}

                                                  //---- setIntRotDevn ---
void molByMoieties::setIntRotDevn(molecGeom &mg) {
  //--- set deviation like vibrational modes for internal rotor
  int iax;
  double angv;
  coordCart vcom;
  devnInfo unitRot;
  (*this)[0].setRotDevn((*this)[1].geom[0], intRotDevn);
  vcom = mg.geom.getVelCOM(intRotDevn);
  intRotDevn -= vcom;
  for (iax = 0; iax < 3; iax++) {
    angv = mg.geom.getAngVPrcAx(intRotDevn, iax);
    unitRot = mg.geom.getUnitRotPrcAx(iax);
    intRotDevn -= angv * unitRot;
  }
  intRotDevn.normalize();
}

ostream &operator<<(ostream &os, molByMoieties &mbm)
  { return mbm.print(os); }

/*------------------------------------------------------------------------
     global library functions
------------------------------------------------------------------------*/

// --- gaussian-log input functions --------------------------------------

/*----- find gaussian job head ---------------------------------------*/

int findGausJobHead(istream &inf, string &gauv) {
  int found = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 9) == " ********") {
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.substr(0, 9) == " Gaussian")
        { gauv = slin; found = true; }
    } else if (slin.substr(0, 19) == " Link1:  Proceeding") {
      found = true;
    }
  } while (!found);
  return NORM;
}

/*----- get gaussian-job route-section -------------------------------*/

string getGausRouteSect(istream &inf) {
  int endroute = false;
  string slin, route;
  do {
    if (getLine(inf, slin) == ERRE) { return ""; }
    if (slin.substr(0, 6) == " -----") {
      if (getLine(inf, slin) == ERRE) { return ""; }
      if (slin.substr(0, 2) == " #") {
        route = slin;
        do {
          if (getLine(inf, slin) == ERRE) { return ""; }
          if (slin.substr(0, 6) == " -----") { endroute = true; }
          else { route += slin; }
        } while (!endroute);
      }
    }
  } while (!endroute);
  strAllToUpper(route);
  return route;
}

/*----- get gaussian-job comment-line --------------------------------*/

string getGausCommLine(istream &inf) {
  int endcomm = false;
  string slin, comm;
  do {
    if (getLine(inf, slin) == ERRE) { return ""; }
    if (slin.substr(0, 6) == " -----") {
      if (getLine(inf, slin) == ERRE) { return ""; }
      comm = slin;
      do {
        if (getLine(inf, slin) == ERRE) { return ""; }
        if (slin.substr(0, 6) == " -----") { endcomm = true; }
        else { comm += slin; }
      } while (!endcomm);
    }
  } while (!endcomm);
  if (comm.length() > 0) {
    if (comm[0] == ' ') { comm = comm.substr(1); }
  }
  return comm;
}

/*----- get gaussian-job archive-part --------------------------------*/

/*
int getGausArchive(istream &inf, string &arch) {
}
*/

/*----- find gaussian job end ----------------------------------------*/

int findGausJobEnd(istream &inf, string &jcpu) {
  int found = false;
  string slin;
  do {
    if (getLine(inf, slin) == ERRE) { return ERRE; }
    if (slin.substr(0, 14) == " Job cpu time:") {
      jcpu = slin;
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (getLine(inf, slin) == ERRE) { return ERRE; }
      if (slin.find("termination") != string::npos) { found = true; }
    }
  } while (!found);
  return NORM;
}

// --- gpop file input / output function ---------------------------------

/*----- read gpop file into objects ------------------------------------*/

int readGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft) {
  string gpofln = basfln + ".gpo";
  ifstream inf;
  inf.open(gpofln.c_str());
  if (!inf)
    { cout << "failed to open [" << gpofln << "].\n"; return ERRE; }
  if (ji.restoreFrom(inf) == ERRE)
    { cout << "Error reading jobInfos."; inf.close(); return ERRE; }
  if (mg.restoreFrom(inf) == ERRE)
    { cout << "Error reading molecGeom."; inf.close(); return ERRE; }
  if (ft.restoreFrom(inf) == ERRE)
    { cout << "Error reading freqTable."; inf.close(); return ERRE; }
  inf.close();
  return NORM;
}

/*----- get a line and check for matching ------------------------------*/

int getlGpoMatch(istream &inStrm, const string &ref) {
  string sLine;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  if (sLine != ref) { return false; }
  return true;
}

/*----- get a string variable in a line --------------------------------*/

int getlGpoStrv(istream &inStrm, string &vStr) {
  string sLine;
  long p;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  if ((p = sLine.find(' ')) == string::npos) { vStr =""; return true; }
  vStr = sLine.substr(p + 1);
  return true;
}

/*----- get an integer variable in a line ------------------------------*/

int getlGpoIntv(istream &inStrm, int &vInt) {
  string sLine;
  long p;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  if ((p = sLine.find(' ')) == string::npos) { return false; }
  vInt = atoi(sLine.substr(p + 1).c_str());
  return true;
}

/*----- get a double variable in a line --------------------------------*/

int getlGpoDblv(istream &inStrm, double &vDbl) {
  string sLine;
  long p;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  if ((p = sLine.find(' ')) == string::npos) { return false; }
  vDbl = atof(sLine.substr(p + 1).c_str());
  return true;
}

/*----- get multiple double variables in a line ------------------------*/

int getlGpoDblvMlt(istream &inStrm, double mvDbl[], int nVals) {
  int iv;
  string sLine;
  vector<string> tkns;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  tkns = splitTokens(sLine, " ");
  for (iv = 0; iv < nVals; iv++) {
    mvDbl[iv] = atof(tkns[iv + 1].c_str());
  }
  return true;
}

/*----- get a boolean variable in a line -------------------------------*/

int getlGpoBlnv(istream &inStrm, int &vBln) {
  string sLine, s;
  long p;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  if ((p = sLine.find(' ')) == string::npos) { return false; }
  s = sLine.substr(p + 1); strAllToUpper(s);
  if (s == "TRUE") { vBln = true; } else { vBln = false; }
  return true;
}

/*----- get a chemical formula -----------------------------------------*/

int getlGpoChmFrm(istream &inStrm, map<int, int> &cf) {
  int nel, iel, atmcNbr, numAtms;
  string sLine;
  vector<string> tkns;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  tkns = splitTokens(sLine, " ");
  nel = (tkns.size() - 1) / 2;
  cf.clear();
  for (iel = 0; iel < nel; iel++) {
    atmcNbr = elm.getAtmcNbr(tkns[iel * 2 + 1]);
    numAtms = atoi(tkns[iel * 2 + 2].c_str());
    cf.insert(make_pair(atmcNbr, numAtms));
  }
  return true;
}

/*----- get an atomic coordinate ---------------------------------------*/

int getlGpoAtmCoord(istream &inStrm, atomCoord &ac) {
  int ixyz;
  string sLine;
  vector<string> tkns;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  tkns = splitTokens(sLine, " ");
  ac.atomicNumber = atoi(tkns[0].c_str());
  ac.atomicType = atoi(tkns[1].c_str());
  for (ixyz = 0; ixyz < 3; ixyz++) {
    ac[ixyz] = atof(tkns[ixyz + 2].c_str());
  }
  return true;
}

/*----- get an atomic coordinate (ambiguous format) --------------------
  number of
  tokens     format
    4            aSym      x y z
    5            aNbr aTyp x y z  (gpop format)
    6        idx aNbr aTyp x y z
                                                          --------------*/
int getlAmbAtmCoord(istream &inStrm, atomCoord &ac) {
  int ixyz, ofsxyz;
  string sLine;
  vector<string> tkns;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  tkns = splitTokens(sLine, " ");
  switch (tkns.size()) {
    case 4:
      ac.atomicNumber = elm.getAtmcNbr(tkns[0]);
      ac.atomicType = 0;
      ofsxyz = 1;
      break;
    case 5:
      ac.atomicNumber = atoi(tkns[0].c_str());
      ac.atomicType = atoi(tkns[1].c_str());
      ofsxyz = 2;
      break;
    case 6:
      ac.atomicNumber = atoi(tkns[1].c_str());
      ac.atomicType = atoi(tkns[2].c_str());
      ofsxyz = 3;
      break;
    default: return false;
  }
  for (ixyz = 0; ixyz < 3; ixyz++) {
    ac[ixyz] = atof(tkns[ixyz + ofsxyz].c_str());
  }
  return true;
}

/*----- get a frequency information ------------------------------------*/

int getlGpofrqInfo(istream &inStrm, freqInfo &fi) {
  string sLine;
  vector<string> tkns;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  tkns = splitTokens(sLine, " ");
  fi.sym = tkns[0]; fi.freq = atof(tkns[1].c_str());
  fi.rmass = atof(tkns[2].c_str()); fi.fconst = atof(tkns[3].c_str());
  fi.irint = atof(tkns[4].c_str()); fi.ramact = atof(tkns[5].c_str());
  fi.depol = atof(tkns[6].c_str());
  return true;
}

/*----- get a Cartesian coordinate -------------------------------------*/

int getlGpoCrdCart(istream &inStrm, coordCart &crd) {
  int ixyz;
  string sLine;
  vector<string> tkns;
  if (getLineGpop(inStrm, sLine) == ERRE) { return false; }
  tkns = splitTokens(sLine, " ");
  for (ixyz = 0; ixyz < 3; ixyz++) {
    crd[ixyz] = atof(tkns[ixyz].c_str());
  }
  return true;
}

/*----- get a key followed by tokens -----------------------------------*/

int getlGpoKeyToks(istream &inStrm, string &key, vector<string> &toks) {
  int itk, ntk;
  string slin;
  vector<string> tkns;
  if (getLineGpop(inStrm, slin) == ERRE) { return false; }
  tkns = splitTokens(slin, " ");
  key = tkns[0]; ntk = tkns.size(); toks.clear();
  for (itk = 1; itk < ntk; itk++) { toks.push_back(tkns[itk]); }
  return true;
}

/*----- get a line from stream removing comment part (#!) --------------*/

int getLineGpop(istream &inStrm, string &sLine) {
  return getLineRmCmm(inStrm, sLine, "#!");
}

/*----- get a boolean value from a string ------------------------------*/

int getBlnvStr(string s) {
  int retv;
  strAllToUpper(s);
  if (s == "TRUE") { retv = true; } else { retv = false; }
  return retv;
}

/*----- output .gpo file -----------------------------------------------*/

int outFileGpo(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, string &gaufln) {
  int prec;
  double eSCFzpe = atof(ji.SCFenergy.c_str()) + ft.zpe / WAVNHRT;
  string gpofln;
  ofstream gpof;
  gpofln = basfln + ".gpo";
  gpof.open(gpofln.c_str());
  if (!gpof)
    { cout << "failed to open [" << gpofln << "].\n"; return ERRE; }
  gpof << "# gpop link data file created from " << gaufln << ".\n";
  prec = gpof.precision(); gpof.setf(ios::fixed); gpof.precision(10);
  gpof << "# SCFenergy + zpe = " << eSCFzpe << endl;
  gpof.unsetf(ios::fixed); gpof.precision(prec);
  gpof << ji << mg << ft;
  gpof.close();
  return NORM;
}

/*----- output .tst file -----------------------------------------------*/

//----- create .tst file only from Gaussian output -------------

int outFileTst(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft) {
  int irot, ifrq, isTS = false, ifrqStart = 0, prec;
  string toutfln;
  ofstream tof;

  toutfln = basfln + ".tst";
  tof.open(toutfln.c_str());
  if (!tof)
    { cout << "failed to open [" << toutfln << "].\n"; return ERRE; }
  prec = tof.precision(); tof.setf(ios::fixed);
  tof << "[Name of the Molecule]\n";
  tof << " " << ji.comment << endl;
  tof << "[Molecular Mass(amu)]\n";
  tof.precision(6); tof << " " << mg.molW << endl;
  tof << "[Deg Electron State |neg|=NumState followed by (gi, ei(cm-1))'s]\n";
  tof << " " << mg.gElec << endl;
  tof << "[Rot Sym Num / Num of Isomrs]\n";
  tof.precision(10); tof << " " << mg.rotSymNbr / mg.numIsomers << endl;
  tof << "[Num of RotConsts (1:linear 3:nonlinear), RotConsts(GHz)]\n";
  tof.precision(7);
  if (mg.nAtms == 1) {
    tof << " 0\n";
  } else {
    if (mg.isLinear) {
      tof << " 1\n";
      tof << " " << mg.rotcGHz[1] << endl;
    } else {
      tof << " 3\n";
      for (irot = 0; irot < 3; irot++)
        { tof << " " << mg.rotcGHz[irot] << endl; }
    }
  }
  if (ft[0].freq < 0.) { isTS = true; ifrqStart = 1; }
  tof << "[Num of Freqs, Freqs(cm-1) |neg|=1D-rotorB(cm-1)]\n";
  if (isTS) { tof << " " << (mg.nVibm - 1) << endl; }
  else { tof << " " << mg.nVibm << endl; }
  tof.precision(4);
  for (ifrq = ifrqStart; ifrq < mg.nVibm; ifrq++) {
    tof << " " << ft[ifrq].freq * ji.scaleFactVib << endl;
  }
  tof << "[Num of H.R. / B(cm-1), Freq(cm-1), SigN]\n";
  tof << " 0\n";
  tof << "[energy for TST (hartree)]\n";
  tof.precision(8); tof << " " << 0. << endl;
  if (isTS) {
    tof << "[imaginary frequency (cm-1) .. only for TS molecule]\n";
    tof.precision(4); tof << " " << fabs(ft[0].freq) << endl;
  }
  tof.unsetf(ios::fixed); tof.precision(prec);
  tof.close();
  return NORM;
}

/*----- prepare a string of boolean variable ---------------------------*/

string strGpoBlnv(int &vBln) {
  string rets;
  if (vBln) { rets = "true"; } else { rets = "false"; }
  return rets;
}

/*------------------------------------------------------------------------
     Declaration of globals
------------------------------------------------------------------------*/

tableOfElements elm;
tableOfMethods gmth;

