/* ======================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
 ====================================================================== */

#ifndef _GPOPLIBU_H_
#define _GPOPLIBU_H_

#include "gpoplibt.h"

/*------------------------------------------------------------------------
     Buffer gas properties
------------------------------------------------------------------------*/

class bufMolec {
 public:                             // --- data members ---
  string name;
  double molW;          // molecular weight [amu]
  double sigma;         // collision diameter [angstrom]
  double epsK;          // collision well depth [K]
 public:                         // --- member functions ---
  bufMolec();
  void clear();
};

class tableBufMol : public vector<bufMolec> {
 public:                         // --- member functions ---
  tableBufMol();
  bufMolec getByName(string n);
};

/*------------------------------------------------------------------------
     RRKM calculation properties
------------------------------------------------------------------------*/

class channelInfo {
 public:                             // --- data members ---
  molInfoSet mTS, mProd;
  int flagTunnel;
 public:                         // --- member functions ---
  channelInfo();
  void clear();
};

class rrkmInfo {
 public:                             // --- data members ---
  enum { Invalid = 0, Energy = 1, File = 2, ImagFreq = 4 };
  string baseName;
  string title;
  double enCnvFac;            // energy conversion factor
  int gsize;                  // grain size [cm-1]
  int align;                  // flag for grain alignment (true/false)
  double aBaseE;              // base energy for grain alignment
  int offset;
  tempList temperat;
  bufMolec buffer, molec, bmpair;
  molInfoSet mReac;
  vector<channelInfo> channel;
  vector<int> chNbr;
 public:                         // --- member functions ---
  rrkmInfo();
  void clear();
  int readFrom(string basfln);
  int readProp();
  int outRRKM(string basfln);
};                                        // --- globals ---
double roundIntOf(double val, double div);
double roundHalfIntOf(double val, double div);

/*------------------------------------------------------------------------
     Declaration of globals
------------------------------------------------------------------------*/

extern tableBufMol bufm;

#endif

