/*========================================================================
     "introt"  Calcutate rotational constants for internal rotors

             Copyright (c) 2002-2006 by A. Miyoshi, Univ. Tokyo
                                     created: Aug.   4, 2006
             (based on Oct. 26, 2005 version of "gpop6mrt.cc")
                                      edited: Aug.  23, 2006
                                      edited: Nov.  19, 2007
--------------------------------------------------------------------------
 usage: introt basename
--------------------------------------------------------------------------
 - reads basename.inp.
 - output is made to console (standard output).
 - .inp file should contain 1) struncture of the molecule as:
     nAtms number_of_atoms
     geom{
       atomic_number type x y z (in angstrome)
       ...
     }geom
   followed by 2) internal rotor specifications as:
     idxCntPiv:idxPiv-idx1-idx2-...
--------------------------------------------------------------------------
 [Example]
========================================================================*/

#include "gpoplibr.h"

/*===== main function ==================================================*/

int main(int argc, char *argv[]) {
  int prec, nTops, irt, icr, anbr;
  double awgt, orgwgt, chkwgt;
  ios::fmtflags flgs;
  ifstream inf;
  string basfln, inpfln, slin;
  vector<string> dtops, vs;
  geomInfo geom;
  molecWithTops moltop;

  showRevision();
  prec = cout.precision(); flgs = cout.flags();
  cout.setf(ios::fixed | ios::right);
  if (argc != 2) { cout << "usage: introt basename\n"; return 1; }
  basfln = argv[1]; inpfln = basfln + ".inp";
    // --- read input (.inp) ---
  inf.open(inpfln.c_str());
  if (!inf)
    { cout << "failed to open [" << inpfln << "].\n"; return 1; }
  if (geom.restoreFrom(inf) == ERRE)
    { cout << "invalid geometry input.\n"; return 1; }
  while (getLineGpop(inf, slin) == NORM) {
    if (slin.substr(0, 7) == "isotope") {
      vs = splitTokens(slin, " ");
      anbr = atoi(vs[1].c_str()); awgt = atof(vs[2].c_str());
      if ((orgwgt = elm.getAtmcWeight(anbr)) < 0.01) {
        cout << "Ignored invalid atomic number [" << anbr
             << "] in isotope statement.\n";
      } else {
        elm.setAtmcWeight(anbr, awgt); chkwgt = elm.getAtmcWeight(anbr);
        cout << "Atomic weight for atomic number = " << anbr
             << " was changed from " << orgwgt << " to " << chkwgt
             << ".\n";
        geom.setPrcAxesMomInert();
      }
    } else {
      dtops.push_back(slin);
    }
  }
  inf.close();
  cout << "----- Input Geometry -----\n";
  cout << geom;
    // --- set molecWithTops object
  moltop.setGeom(geom);
  if (moltop.setTops(dtops) == ERRE)
    { cout << "invalid rotor description.\n"; return 1; }
  nTops = moltop.aTops.size();
  cout << "----- Uncoupled Properties -----\n";
  cout.precision(5);
  for (irt = 0; irt < nTops; irt++) {
    cout << " ucR#" << irt + 1 << ": I=";
    cout.width(9); cout << moltop.aTops[irt].I0 << " B=";
    cout.width(9); cout << moltop.aTops[irt].rotC0;
    cout << endl;
  }
    // --- coupled rotor calculation
  moltop.calcCoupling();
  cout << "----- Coupling Calculation Results -----\n";
  for (icr = 0; icr < nTops; icr++) {
    cout.precision(5);
    cout << " cpR#" << icr + 1 << ": I=";
    cout.width(9); cout << moltop.moI[icr] << " B=";
    cout.width(9); cout << moltop.roB[icr] << " vec=";
    cout.precision(3);
    for (irt = 0; irt < nTops; irt++)
      { cout << " "; cout.width(6); cout << moltop.coefCpl[icr][irt]; }
    cout << endl;
  }
  cout.flags(flgs);
  cout.precision(prec);
  return 0;
}

